#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/LocalBasisCache.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct testvec_gradtrial {};
  }


  /// first-order operator \f$ \langle\Psi, c\,\nabla\phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::testvec_gradtrial, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::testvec_gradtrial, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::testvec_gradtrial, GridFct const& expr)
      : Super(expr, 1)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isPower && ColNode::isLeaf,
        "ColNode must be Leaf-Node and RowNode must be a Power-Node.");

      static const std::size_t CHILDREN = RowNode::CHILDREN;
      static_assert( CHILDREN == Context::dow, "" );

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;

      using RangeFieldType = typename ColLocalBasisType::Traits::RangeFieldType;

      LocalBasisCache<RowLocalBasisType> rowCache(rowLocalFE.localBasis());
      LocalBasisCache<ColLocalBasisType> colCache(colLocalFE.localBasis());

      auto const& shapeValuesCache = rowCache.evaluateFunctionAtQp(context, quad);
      auto const& shapeGradientsCache = colCache.evaluateJacobianAtQp(context, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        // the values of the shape functions on the reference element at the quadrature point
        auto const& shapeValues = shapeValuesCache[iq];

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using WorldVector = FieldVector<RangeFieldType,Context::dow>;
        thread_local std::vector<WorldVector> colGradients;
        colGradients.resize(shapeGradients.size());

        for (std::size_t i = 0; i < colGradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], colGradients[i]);

        for (std::size_t i = 0; i < rowSize; ++i) {
          const auto value = factor * shapeValues[i];
          for (std::size_t j = 0; j < colSize; ++j) {
            const auto local_j = colNode.localIndex(j);
            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = rowNode.child(k).localIndex(i);
              elementMatrix[local_ki][local_j] += value * colGradients[j][k];
            }
          }
        }
      }

    }
  };

  /** @} **/

} // end namespace AMDiS
