#pragma once

#include <amdis/Assembler.hpp>
#include <amdis/LocalAssembler.hpp>
#include <amdis/LocalOperator.hpp>
#include <amdis/utility/Visitor.hpp>

namespace AMDiS {

template <class BS, class B>
void DOFVectorBase<BS,B>::
init(bool asmVector)
{
  backend_.resize(size());
  if (asmVector)
    backend_.set(0);

  auto const localSize = basis_->localView().maxSize();
  elementVector_.resize(localSize);
}


template <class BS, class B>
void DOFVectorBase<BS,B>::
finish(bool /*asmVector*/)
{}


template <class BS, class B>
void DOFVectorBase<BS,B>::
insert(LocalView const& localView, ElementVector const& elementVector)
{
  for (size_type i = 0; i < localView.size(); ++i) {
    if (std::abs(elementVector[i]) > threshold<double>) {
      size_type const idx = flatMultiIndex(localView.index(i));
      backend_[idx] += elementVector[i];
    }
  }
}


template <class BS, class B>
  template <class ContextTag, class Operator, class TreePath>
void DOFVectorBase<BS,B>::
addOperator(ContextTag contextTag, Operator const& preOp, TreePath path)
{
  static_assert( Concepts::PreTreePath<TreePath>,
      "path must be a valid treepath, or an integer/index-constant");

  auto i = child(basis_->localView().tree(), makeTreePath(path));

  using Context = typename ContextTag::type;
  auto op = makeLocalOperator<Context>(preOp, basis_->gridView());
  auto localAssembler = makeLocalAssemblerPtr<Context>(std::move(op), i);

  operators_[i].push(contextTag, localAssembler);
}


template <class BS, class B>
void DOFVectorBase<BS,B>::
assemble(LocalView const& localView)
{
  elementVector_ = 0;
  auto const& gv = basis_->gridView();
  auto const& element = localView.element();
  auto geometry = element.geometry();

  forEachNode_(localView.tree(), [&](auto const& node, auto) {
    auto& rhsOp = operators_[node];
    if (rhsOp) {
      rhsOp.bind(element, geometry);
      assembleOperators(gv, element, rhsOp, [&](auto const& context, auto& operators) {
        for (auto scaled : operators)
          scaled.op->assemble(context, node, elementVector_);
      });
      rhsOp.unbind();
    }
  });

  insert(localView, elementVector_);
}


template <class BS, class B>
void DOFVectorBase<BS,B>::
assemble()
{
  auto localView = basis_->localView();

  init(true);
  for (auto const& element : elements(basis_->gridView())) {
    localView.bind(element);
    assemble(localView);
    localView.unbind(element);
  }
  finish(true);
}

} // end namespace AMDiS
